@@ -21,11 +21,47 @@
 #ifndef COMMON_H
 #define COMMON_H
 
+/**************************************** Common constants */
+
+#ifndef VERSION
+#define VERSION "4.17"
+#endif
+
+#ifdef OPTIMIZE_SCALABILITY
+
+/* CPU stack size */
+#define STACK_SIZE      32768
+#define DEBUG_STACK_SIZE
+
+/* I/O buffer size */
+#define BUFFSIZE        2048
+
+/* Length of strings (including the terminating '\0' character) */
+#define STRLEN          128
+
+#else /* OPTIMIZE_SCALABILITY */
+
+/* CPU stack size */
+#define STACK_SIZE      65536
+
+/* I/O buffer size */
+#define BUFFSIZE        16384
+
+/* Length of strings (including the terminating '\0' character) */
+#define STRLEN          256
+
+#endif /* OPTIMIZE_SCALABILITY */
+
+/* IP address and TCP port textual representation length */
+#define IPLEN           128
+
+/* How many bytes of random input to read from files for PRNG */
+/* OpenSSL likes at least 128 bits, so 64 bytes seems plenty. */
+#define RANDOM_BYTES    64
+
 /* For FormatGuard */
 /* #define __NO_FORMATGUARD_ */
 
-/* #define DEBUG_STACK_SIZE */
-
 /**************************************** Platform */
 
 #ifdef USE_WIN32
@@ -44,28 +80,6 @@ typedef int socklen_t;
 #define HAVE_OSSL_ENGINE_H
 #endif
 
-/**************************************** Common constants */
-
-#ifndef VERSION
-#define VERSION "4.15"
-#endif
-
-/* CPU stack size */
-#define STACK_SIZE 65536
-
-/* I/O buffer size */
-#define BUFFSIZE        16384
-
-/* Length of strings (including the terminating '\0' character) */
-#define STRLEN          256
-
-/* IP address and TCP port textual representation length */
-#define IPLEN           128
-
-/* How many bytes of random input to read from files for PRNG */
-/* OpenSSL likes at least 128 bits, so 64 bytes seems plenty. */
-#define RANDOM_BYTES    64
-
 /**************************************** Generic headers */
 
 #ifdef __vms
@@ -188,13 +202,21 @@ typedef unsigned int u32;
 typedef unsigned long u32;
 #endif
 
+#ifdef __INNOTEK_LIBC__
+# define get_last_socket_error() sock_errno()
+# define get_last_error()        errno
+# define readsocket(s,b,n)       recv((s),(b),(n),0)
+# define writesocket(s,b,n)      send((s),(b),(n),0)
+# define closesocket(s)          close(s)
+# define ioctlsocket(a,b,c)      so_ioctl((a),(b),(c))
+#else
 #define get_last_socket_error() errno
 #define get_last_error()        errno
 #define readsocket(s,b,n)       read((s),(b),(n))
 #define writesocket(s,b,n)      write((s),(b),(n))
 #define closesocket(s)          close(s)
 #define ioctlsocket(a,b,c)      ioctl((a),(b),(c))
-
+#endif
     /* OpenVMS compatibility */
 #ifdef __vms
 #define libdir "__NA__"
@@ -296,12 +318,15 @@ extern char *sys_errlist[];
 #ifdef HAVE_OSSL_ENGINE_H
 #include <openssl/engine.h>
 #endif
-#else
+#if SSLEAY_VERSION_NUMBER >= 0x00907000L
+#include <openssl/ocsp.h>
+#endif /* OpenSSL-0.9.7 */
+#else /* HAVE_OPENSSL */
 #include <lhash.h>
 #include <ssl.h>
 #include <err.h>
 #include <crypto.h> /* for CRYPTO_* and SSLeay_version */
-#endif
+#endif /* HAVE_OPENSSL */
 
 /**************************************** Other defines */
 
@@ -312,10 +337,12 @@ extern char *sys_errlist[];
     (dst[STRLEN-1]='\0', strncat((dst), (src), STRLEN-strlen(dst)-1))
 /* change all non-printable characters to '.' */
 #define safestring(s) \
-    do {char *p; for(p=(s); *p; p++) if(!isprint((int)*p)) *p='.';} while(0)
+    do {unsigned char *p; for(p=(s); *p; p++) \
+        if(!isprint((int)*p)) *p='.';} while(0)
 /* change all unsafe characters to '.' */
 #define safename(s) \
-    do {char *p; for(p=(s); *p; p++) if(!isalnum((int)*p)) *p='.';} while(0)
+    do {unsigned char *p; for(p=(s); *p; p++) \
+        if(!isalnum((int)*p)) *p='.';} while(0)
 
 /* Some definitions for IPv6 support */
 #if defined(USE_IPv6)
