@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (c) 1998-2005 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (c) 1998-2006 Michal Trojnara <Michal.Trojnara@mirt.net>
  *                 All Rights Reserved
  *
  *   This program is free software; you can redistribute it and/or modify
@@ -31,19 +31,26 @@
 #include "common.h"
 #include "prototypes.h"
 
-#define DEBUG_UCONTEXT
+/* #define DEBUG_UCONTEXT */
 
 #ifndef USE_WIN32
 static int signal_pipe[2]={-1, -1};
 static char signal_buffer[16];
 static void sigchld_handler(int);
+#ifdef __INNOTEK_LIBC__
+struct sockaddr_un {
+    u_char  sun_len;             /* sockaddr len including null */
+    u_char  sun_family;          /* AF_OS2 or AF_UNIX */
+    char    sun_path[108];       /* path name */
+};
+#endif
 static void signal_pipe_empty(void);
 #ifdef USE_FORK
 static void client_status(void); /* dead children detected */
 #endif
 #endif /* !defined(USE_WIN32) */
 
-static void setnonblock(int, unsigned long);
+/**************************************** s_poll functions */
 
 #ifdef USE_POLL
 
@@ -193,7 +200,8 @@ static void scan_waiting_queue(void) {
     }
 }
 
-int s_poll_wait(s_poll_set *fds, int timeout) {
+int s_poll_wait(s_poll_set *fds, int sec, int msec) {
+    /* FIXME: msec parameter is currently ignored with UCONTEXT threads */
     CONTEXT *ctx; /* current context */
     static CONTEXT *to_free=NULL; /* delayed memory deallocation */
 
@@ -205,7 +213,7 @@ int s_poll_wait(s_poll_set *fds, int timeout) {
 
     if(fds) { /* something to wait for -> swap the context */
         ctx->fds=fds; /* set file descriptors to wait for */
-        ctx->finish=timeout<0 ? -1 : time(NULL)+timeout;
+        ctx->finish=sec<0 ? -1 : time(NULL)+sec;
         /* move (append) the current context to the waiting queue */
         ctx->next=NULL;
         if(waiting_tail)
@@ -246,13 +254,13 @@ int s_poll_wait(s_poll_set *fds, int timeout) {
 
 #else /* USE_UCONTEXT */
 
-int s_poll_wait(s_poll_set *fds, int timeout) {
+int s_poll_wait(s_poll_set *fds, int sec, int msec) {
     int retval, retry;
 
     do { /* skip "Interrupted system call" errors */
         retry=0;
-        retval=poll(fds->ufds, fds->nfds, timeout<0 ? -1 : 1000*timeout);
-        if(timeout<0 && retval>0 && s_poll_canread(fds, signal_pipe[0])) {
+        retval=poll(fds->ufds, fds->nfds, sec<0 ? -1 : 1000*sec+msec);
+        if(sec<0 && retval>0 && s_poll_canread(fds, signal_pipe[0])) {
             signal_pipe_empty(); /* no timeout -> main loop */
             retry=1;
         }
@@ -287,7 +295,7 @@ int s_poll_canwrite(s_poll_set *fds, int fd) {
     return FD_ISSET(fd, &fds->owfds);
 }
 
-int s_poll_wait(s_poll_set *fds, int timeout) {
+int s_poll_wait(s_poll_set *fds, int sec, int msec) {
     int retval, retry;
     struct timeval tv, *tv_ptr;
 
@@ -295,16 +303,16 @@ int s_poll_wait(s_poll_set *fds, int timeout) {
         retry=0;
         memcpy(&fds->orfds, &fds->irfds, sizeof(fd_set));
         memcpy(&fds->owfds, &fds->iwfds, sizeof(fd_set));
-        if(timeout<0) { /* infinite timeout */
+        if(sec<0) { /* infinite timeout */
             tv_ptr=NULL;
         } else {
-            tv.tv_sec=timeout;
-            tv.tv_usec=0;
+            tv.tv_sec=sec;
+            tv.tv_usec=1000*msec;
             tv_ptr=&tv;
         }
         retval=select(fds->max+1, &fds->orfds, &fds->owfds, NULL, tv_ptr);
-#ifndef USE_WIN32
-        if(timeout<0 && retval>0 && s_poll_canread(fds, signal_pipe[0])) {
+#if !defined(USE_WIN32) && !defined(USE_OS2)
+        if(sec<0 && retval>0 && s_poll_canread(fds, signal_pipe[0])) {
             signal_pipe_empty(); /* no timeout -> main loop */
             retry=1;
         }
@@ -328,17 +336,56 @@ static void sigchld_handler(int sig) { /* SIGCHLD detected */
     while(wait_for_pid(-1, &status, WNOHANG)>0) {
         /* no logging is possible in a signal handler */
 #ifdef USE_FORK
-        num_clients--; /* one client less */
+        --num_clients; /* one client less */
 #endif /* USE_FORK */
     }
 #else /* __sgi */
+#ifdef __INNOTEK_LIBC__
+    writesocket(signal_pipe[1], signal_buffer, 1);
+#else
     write(signal_pipe[1], signal_buffer, 1);
+#endif /* __INNOTEK_LIBC__ */
 #endif /* __sgi */
     signal(SIGCHLD, sigchld_handler);
     errno=save_errno;
 }
 
+/**************************************** signal pipe */
+
 int signal_pipe_init(void) {
+#if defined(__INNOTEK_LIBC__)
+    /* Innotek port of GCC can not use select on a pipe, use local socket instead */
+    struct sockaddr_un un;
+    fd_set set_pipe;
+    int pipe_in;
+
+    FD_ZERO(&set_pipe);
+    signal_pipe[0] = socket(PF_OS2, SOCK_STREAM, 0);
+    pipe_in=signal_pipe[0];
+    signal_pipe[1] = socket(PF_OS2, SOCK_STREAM, 0);
+
+    alloc_fd(signal_pipe[0]);
+    alloc_fd(signal_pipe[1]);
+
+    /* Connect the two endpoints */
+    memset(&un, 0, sizeof(un));
+
+    un.sun_len = sizeof(un);
+    un.sun_family = AF_OS2;
+    sprintf(un.sun_path, "\\socket\\stunnel-%u", getpid());
+    /* Make the first endpoint listen */
+    bind(signal_pipe[0], (struct sockaddr *)&un, sizeof(un));
+    listen(signal_pipe[0], 5);
+    connect(signal_pipe[1], (struct sockaddr *)&un, sizeof(un));
+    FD_SET(signal_pipe[0], &set_pipe);
+    if (select(signal_pipe[0]+1, &set_pipe, NULL, NULL, NULL)>0) {
+        signal_pipe[0]=accept(signal_pipe[0], NULL, 0);
+        closesocket(pipe_in);
+    } else {
+        sockerror("select");
+        exit(1);
+    }
+#else /* __INNOTEK_LIBC__ */
     if(pipe(signal_pipe)) {
         ioerror("pipe");
         exit(1);
@@ -349,14 +396,19 @@ int signal_pipe_init(void) {
     /* close the pipe in child execvp */
     fcntl(signal_pipe[0], F_SETFD, FD_CLOEXEC);
     fcntl(signal_pipe[1], F_SETFD, FD_CLOEXEC);
-#endif
+#endif /* FD_CLOEXEC */
+#endif /* __INNOTEK_LIBC__ */
     signal(SIGCHLD, sigchld_handler);
     return signal_pipe[0];
 }
 
 static void signal_pipe_empty(void) {
     s_log(LOG_DEBUG, "Cleaning up the signal pipe");
+#ifdef __INNOTEK_LIBC__
+    readsocket(signal_pipe[0], signal_buffer, sizeof(signal_buffer));
+#else
     read(signal_pipe[0], signal_buffer, sizeof(signal_buffer));
+#endif
 #ifdef USE_FORK
     client_status(); /* report status of client process */
 #else /* USE_UCONTEXT || USE_PTHREAD */
@@ -370,10 +422,10 @@ static void client_status(void) { /* dead children detected */
 
 #ifdef HAVE_WAIT_FOR_PID
     while((pid=wait_for_pid(-1, &status, WNOHANG))>0) {
-        num_clients--; /* one client less */
+        --num_clients; /* one client less */
 #else
     if((pid=wait(&status))>0) {
-        num_clients--; /* one client less */
+        --num_clients; /* one client less */
 #endif
 #ifdef WIFSIGNALED
         if(WIFSIGNALED(status)) {
@@ -417,6 +469,8 @@ void child_status(void) { /* dead libwrap or 'exec' process detected */
 
 #endif /* !defined USE_WIN32 */
 
+/**************************************** fd management */
+
 int alloc_fd(int sock) {
 #ifndef USE_WIN32
     if(!max_fds || sock>=max_fds) {
@@ -435,8 +489,8 @@ int alloc_fd(int sock) {
 #define O_NONBLOCK O_NDELAY
 #endif
 
-static void setnonblock(int sock, unsigned long l) {
-#if defined F_GETFL && defined F_SETFL && defined O_NONBLOCK
+void setnonblock(int sock, unsigned long l) {
+#if defined F_GETFL && defined F_SETFL && defined O_NONBLOCK && !defined __INNOTEK_LIBC__
     int retval, flags;
     do {
         flags=fcntl(sock, F_GETFL, 0);
@@ -486,145 +540,165 @@ int set_socket_options(int s, int type) {
     return 0; /* OK */
 }
 
-int write_blocking(CLI *c, int fd, u8 *ptr, int len) {
+/**************************************** simulate blocking I/O */
+
+void write_blocking(CLI *c, int fd, void *ptr, int len) {
         /* simulate a blocking write */
-        /* returns 0 on success, -1 on failure */
     s_poll_set fds;
     int num;
 
     while(len>0) {
         s_poll_zero(&fds);
         s_poll_add(&fds, fd, 0, 1); /* write */
-        switch(s_poll_wait(&fds, c->opt->timeout_busy)) {
+        switch(s_poll_wait(&fds, c->opt->timeout_busy, 0)) {
         case -1:
             sockerror("write_blocking: s_poll_wait");
-            return -1; /* error */
+            longjmp(c->err, 1); /* error */
         case 0:
             s_log(LOG_INFO, "write_blocking: s_poll_wait timeout");
-            return -1; /* timeout */
+            longjmp(c->err, 1); /* timeout */
         case 1:
             break; /* OK */
         default:
             s_log(LOG_ERR, "write_blocking: s_poll_wait unknown result");
-            return -1; /* error */
+            longjmp(c->err, 1); /* error */
         }
         num=writesocket(fd, ptr, len);
         switch(num) {
         case -1: /* error */
             sockerror("writesocket (write_blocking)");
-            return -1;
+            longjmp(c->err, 1);
         }
-        ptr+=num;
+        ptr=(u8 *)ptr+num;
         len-=num;
     }
-    return 0; /* OK */
 }
 
-int read_blocking(CLI *c, int fd, u8 *ptr, int len) {
+void read_blocking(CLI *c, int fd, void *ptr, int len) {
         /* simulate a blocking read */
-        /* returns 0 on success, -1 on failure */
     s_poll_set fds;
     int num;
 
     while(len>0) {
         s_poll_zero(&fds);
         s_poll_add(&fds, fd, 1, 0); /* read */
-        switch(s_poll_wait(&fds, c->opt->timeout_busy)) {
+        switch(s_poll_wait(&fds, c->opt->timeout_busy, 0)) {
         case -1:
             sockerror("read_blocking: s_poll_wait");
-            return -1; /* error */
+            longjmp(c->err, 1); /* error */
         case 0:
             s_log(LOG_INFO, "read_blocking: s_poll_wait timeout");
-            return -1; /* timeout */
+            longjmp(c->err, 1); /* timeout */
         case 1:
             break; /* OK */
         default:
             s_log(LOG_ERR, "read_blocking: s_poll_wait unknown result");
-            return -1; /* error */
+            longjmp(c->err, 1); /* error */
         }
         num=readsocket(fd, ptr, len);
         switch(num) {
         case -1: /* error */
             sockerror("readsocket (read_blocking)");
-            return -1;
+            longjmp(c->err, 1);
         case 0: /* EOF */
             s_log(LOG_ERR, "Unexpected socket close (read_blocking)");
-            return -1;
+            longjmp(c->err, 1);
         }
-        ptr+=num;
+        ptr=(u8 *)ptr+num;
         len-=num;
     }
-    return 0; /* OK */
 }
 
-int fdprintf(CLI *c, int fd, const char *format, ...) {
-    va_list arglist;
-    char line[STRLEN], logline[STRLEN];
-    char crlf[]="\r\n";
+void fdputline(CLI *c, int fd, const char *line) {
+    char tmpline[STRLEN];
+    const char crlf[]="\r\n";
     int len;
 
-    va_start(arglist, format);
-#ifdef HAVE_VSNPRINTF
-    len=vsnprintf(line, STRLEN, format, arglist);
-#else
-    len=vsprintf(line, format, arglist);
-#endif
-    va_end(arglist);
-    len+=2;
-    if(len>=STRLEN) {
-        s_log(LOG_ERR, "Line too long in fdprintf");
-        return -1;
-    }
-    safecopy(logline, line); /* The line without crlf */
-    safeconcat(line, crlf);
-    if(write_blocking(c, fd, line, len)<0)
-        return -1;
-    safestring(logline);
-    s_log(LOG_DEBUG, " -> %s", logline);
-    return len;
+    safecopy(tmpline, line);
+    safeconcat(tmpline, crlf);
+    len=strlen(tmpline);
+    write_blocking(c, fd, tmpline, len);
+    tmpline[len-2]='\0'; /* remove CRLF */
+    safestring(tmpline);
+    s_log(LOG_DEBUG, " -> %s", tmpline);
 }
 
-int fdscanf(CLI *c, int fd, const char *format, char *buffer) {
-    char line[STRLEN], logline[STRLEN], lformat[STRLEN];
+void fdgetline(CLI *c, int fd, char *line) {
+    char tmpline[STRLEN];
     s_poll_set fds;
-    int ptr, retval;
+    int ptr;
 
-    for(ptr=0; ptr<STRLEN-1; ptr++) {
+    for(ptr=0;;) {
         s_poll_zero(&fds);
         s_poll_add(&fds, fd, 1, 0); /* read */
-        switch(s_poll_wait(&fds, c->opt->timeout_busy)) {
+        switch(s_poll_wait(&fds, c->opt->timeout_busy, 0)) {
         case -1:
-            sockerror("fdscanf: s_poll_wait");
-            return -1; /* error */
+            sockerror("fdgetline: s_poll_wait");
+            longjmp(c->err, 1); /* error */
         case 0:
-            s_log(LOG_INFO, "fdscanf: s_poll_wait timeout");
-            return -1; /* timeout */
+            s_log(LOG_INFO, "fdgetline: s_poll_wait timeout");
+            longjmp(c->err, 1); /* timeout */
         case 1:
             break; /* OK */
         default:
-            s_log(LOG_ERR, "fdscanf: s_poll_wait unknown result");
-            return -1; /* error */
+            s_log(LOG_ERR, "fdgetline: s_poll_wait unknown result");
+            longjmp(c->err, 1); /* error */
         }
         switch(readsocket(fd, line+ptr, 1)) {
         case -1: /* error */
-            sockerror("readsocket (fdscanf)");
-            return -1;
+            sockerror("readsocket (fdgetline)");
+            longjmp(c->err, 1);
         case 0: /* EOF */
-            s_log(LOG_ERR, "Unexpected socket close (fdscanf)");
-            return -1;
+            s_log(LOG_ERR, "Unexpected socket close (fdgetline)");
+            longjmp(c->err, 1);
         }
         if(line[ptr]=='\r')
             continue;
         if(line[ptr]=='\n')
             break;
+        if(!line[ptr])
+            break;
+        if(++ptr==STRLEN) {
+            s_log(LOG_ERR, "Input line too long");
+            longjmp(c->err, 1);
+        }
     }
     line[ptr]='\0';
-    safecopy(logline, line);
-    safestring(logline);
-    s_log(LOG_DEBUG, " <- %s", logline);
+    safecopy(tmpline, line);
+    safestring(tmpline);
+    s_log(LOG_DEBUG, " <- %s", tmpline);
+}
+
+int fdprintf(CLI *c, int fd, const char *format, ...) {
+    va_list arglist;
+    char line[STRLEN];
+    int len;
+
+    va_start(arglist, format);
+#ifdef HAVE_VSNPRINTF
+    len=vsnprintf(line, STRLEN-2, format, arglist);
+#else
+    len=vsprintf(line, format, arglist);
+#endif
+    va_end(arglist);
+    if(len<0) {
+        s_log(LOG_ERR, "fdprintf: vs(n)printf failed");
+        longjmp(c->err, 1);
+    }
+    fdputline(c, fd, line);
+    return len+2;
+}
+
+int fdscanf(CLI *c, int fd, const char *format, char *buffer) {
+    char line[STRLEN], lformat[STRLEN];
+    int ptr, retval;
+
+    fdgetline(c, fd, line);
+
     retval=sscanf(line, format, buffer);
     if(retval>=0)
         return retval;
+
     s_log(LOG_DEBUG, "fdscanf falling back to lowercase");
     safecopy(lformat, format);
     for(ptr=0; lformat[ptr]; ptr++)
