@@ -66,7 +66,7 @@
  */
 
 int pty_allocate(int *ptyfd, int *ttyfd, char *namebuf, int namebuflen) {
-#if defined(HAVE_OPENPTY) || defined(BSD4_4)
+#if defined(HAVE_OPENPTY) || defined(BSD4_4) && !defined(__INNOTEK_LIBC__)
     /* openpty(3) exists in OSF/1 and some other os'es */
     char buf[64];
     int i;
