@@ -11,5 +11,7 @@
 #define IDM_CLOSE 41
 #define IDM_SETUP 42
 
-#define IDE_PASSWORDEDIT 50
+#define IDE_EDIT 50
+#define IDE_PASSEDIT 51
+#define IDE_PINEDIT 52
 
