@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (c) 1998-2005 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (c) 1998-2006 Michal Trojnara <Michal.Trojnara@mirt.net>
  *                 All Rights Reserved
  *
  *   This program is free software; you can redistribute it and/or modify
@@ -28,6 +28,11 @@
  *   do so, delete this exception statement from your version.
  */
 
+#ifdef USE_OS2
+#define INCL_DOSPROCESS
+#include <os2.h>
+#endif
+
 #include "common.h"
 #include "prototypes.h"
 
@@ -46,6 +51,21 @@ void leave_critical_section(SECTION_CODE i) {
 
 #ifdef USE_UCONTEXT
 
+#if defined(CPU_SPARC) && ( \
+        defined(OS_SOLARIS2_0) || \
+        defined(OS_SOLARIS2_1) || \
+        defined(OS_SOLARIS2_2) || \
+        defined(OS_SOLARIS2_3) || \
+        defined(OS_SOLARIS2_4) || \
+        defined(OS_SOLARIS2_5) || \
+        defined(OS_SOLARIS2_6) || \
+        defined(OS_SOLARIS2_7) || \
+        defined(OS_SOLARIS2_8))
+#define ARGC 2
+#else
+#define ARGC 1
+#endif
+
 /* first context on the ready list is the active context */
 CONTEXT *ready_head=NULL, *ready_tail=NULL;         /* ready to execute */
 CONTEXT *waiting_head=NULL, *waiting_tail=NULL;     /* waiting on poll() */
@@ -78,9 +98,9 @@ static CONTEXT *new_context(void) {
         return NULL;
     }
     ctx->ctx.uc_link=NULL; /* it should never happen */
-#if defined(__sparc) && !defined(HAVE___MAKECONTEXT_V2)
+#if defined(__sgi) || ARGC==2 /* obsolete ss_sp semantics */
     ctx->ctx.uc_stack.ss_sp=ctx->stack+STACK_SIZE-8;
-#else /* not an old and buggy Solaris on Sparc */
+#else
     ctx->ctx.uc_stack.ss_sp=ctx->stack;
 #endif
     ctx->ctx.uc_stack.ss_size=STACK_SIZE;
@@ -113,7 +133,7 @@ int create_client(int ls, int s, void *arg, void *(*cli)(void *)) {
     if(!ctx)
         return -1;
     s_log(LOG_DEBUG, "Context %ld created", ctx->id);
-    makecontext(&ctx->ctx, (void(*)(void))cli, 1, arg);
+    makecontext(&ctx->ctx, (void(*)(void))cli, ARGC, arg);
     return 0;
 }
 
@@ -299,11 +319,69 @@ int create_client(int ls, int s, void *arg, void *(*cli)(void *)) {
     return 0;
 }
 
+#endif
+ 
+#ifdef USE_OS2
+
+void enter_critical_section(SECTION_CODE i) {
+    DosEnterCritSec();
+}
+
+void leave_critical_section(SECTION_CODE i) {
+    DosExitCritSec();
+}
+
+void sthreads_init(void) {
+}
+
+unsigned long stunnel_process_id(void) {
+    PTIB ptib=NULL;
+    DosGetInfoBlocks(&ptib, NULL);
+    return (unsigned long)ptib->tib_ordinal;
+}
+
+unsigned long stunnel_thread_id(void) {
+    PPIB ppib=NULL;
+    DosGetInfoBlocks(NULL, &ppib);
+    return (unsigned long)ppib->pib_ulpid;
+}
+
+int create_client(int ls, int s, void *arg, void *(*cli)(void *)) {
+    s_log(LOG_DEBUG, "Creating a new thread");
+    if(_beginthread((void(*)(void *))cli, NULL, STACK_SIZE, arg)==-1) {
+        ioerror("_beginthread");
+        return -1;
+    }
+    s_log(LOG_DEBUG, "New thread created");
+    return 0;
+}
+
+#ifdef _WIN32_WCE
+
+int _beginthread(void (*start_address)(void *),
+        int stack_size, void *arglist) {
+    DWORD thread_id;
+    HANDLE handle;
+
+    handle=CreateThread(NULL, stack_size,
+        (LPTHREAD_START_ROUTINE)start_address, arglist, 0, &thread_id);
+    if(!handle)
+        return -1;
+    CloseHandle(handle);
+    return 0;
+}
+
+void _endthread(void) {
+    ExitThread(0);
+}
+
+#endif /* !defined(_WIN32_WCE) */
+
 #endif /* USE_WIN32 */
 
 #ifdef DEBUG_STACK_SIZE
 
-#define STACK_RESERVE (STACK_SIZE/4)
+#define STACK_RESERVE (STACK_SIZE/8)
 #define VERIFY_AREA ((STACK_SIZE-STACK_RESERVE)/sizeof(u32))
 #define TEST_VALUE 0xdeadbeef
 
@@ -329,7 +407,7 @@ void stack_info(int init) { /* 1-initialize, 0-display */
         if(i>num) /* use the higher value */
             num=i;
         if(num<64) {
-            s_log(LOG_ERR, "STACK_RESERVE is too high");
+            s_log(LOG_NOTICE, "STACK_RESERVE is too high");
             return;
         }
         if(num<min_num)
@@ -337,10 +415,10 @@ void stack_info(int init) { /* 1-initialize, 0-display */
         s_log(LOG_NOTICE,
             "stack_info: size=%d, current=%d (%d%%), maximum=%d (%d%%)",
             STACK_SIZE,
-            (VERIFY_AREA-num)*sizeof(u32),
-            (VERIFY_AREA-num)*sizeof(u32)*100/STACK_SIZE,
-            (VERIFY_AREA-min_num)*sizeof(u32),
-            (VERIFY_AREA-min_num)*sizeof(u32)*100/STACK_SIZE);
+            (int)((VERIFY_AREA-num)*sizeof(u32)),
+            (int)((VERIFY_AREA-num)*sizeof(u32)*100/STACK_SIZE),
+            (int)((VERIFY_AREA-min_num)*sizeof(u32)),
+            (int)((VERIFY_AREA-min_num)*sizeof(u32)*100/STACK_SIZE));
     }
 }
 
